﻿using System;
using System.Collections.Generic;

namespace UMK.WzorceOperacyjne
{
    //https://sourcemaking.com/design_patterns/observer/cpp/1

    /*
    interface IObserwator
    {
        void Aktualizuj(int parametr);
    }
    */

    abstract class Obserwator : IObserver<int>
    {
        protected int wartość = 0;
        protected int dzielnik;

        public Obserwator(int dzielnik)
        {
            this.dzielnik = dzielnik;
        }

        public abstract void OnNext(int value);

        public void OnError(Exception exc)
        {
            Console.Error.WriteLine("Błąd obserwatora: + " + exc.Message);
        }

        public void OnCompleted()
        {
            Console.WriteLine("Działanie obserwatora zakończone");
        }
    }

    class Dzielenie : Obserwator
    {
        public Dzielenie(int dzielnik)
            :base(dzielnik)
        {             
        }

        //public void Aktualizuj(int dzielna)
        public override void OnNext(int dzielna)
        {
            wartość = dzielna / dzielnik;
            Console.WriteLine("{0} / {1} = {2}", dzielna, dzielnik, wartość);
        }
    }

    class Modulo : Obserwator
    {
        public Modulo(int dzielnik)
            :base(dzielnik)
        {            
        }

        public override void OnNext(int dzielna)
        {
            wartość = dzielna % dzielnik;
            Console.WriteLine("{0} % {1} = {2}", dzielna, dzielnik, wartość);
        }
    }

    class Unsubscriber : IDisposable
    {
        private List<IObserver<int>> obserwatorzy;
        private IObserver<int> obserwator;

        public Unsubscriber(List<IObserver<int>> obserwatorzy, IObserver<int> obserwator)
        {
            this.obserwatorzy = obserwatorzy;
            this.obserwator = obserwator;
        }

        public void Dispose()
        {
            if (obserwatorzy != null && obserwatorzy.Contains(obserwator))
                obserwatorzy.Remove(obserwator);
        }
    }

    class Podmiot : IObservable<int>
    {
        private int wartość;
        //private Dzielenie dzielenie; //obiekty, których stan zależy od zmieniającego się pola wartość
        //private Modulo modulo;
        private List<IObserver<int>> obserwatorzy = new List<IObserver<int>>();

        /*
        private void powiadom()
        {
            //dzielenie.Aktualizuj(wartość);
            //modulo.Aktualizuj(wartość);
            foreach (IObserwator<int> obserwator in obserwatorzy)
                obserwator.Aktualizuj(wartość);
        }
        */

        //public void PrzyłączObserwator(IObserwator obserwator)
        public IDisposable Subscribe(IObserver<int> obserwator)
        {
            if (!obserwatorzy.Contains(obserwator))
                obserwatorzy.Add(obserwator);
            return new Unsubscriber(obserwatorzy, obserwator);
        }

        public void Zakończ()
        {
            foreach (IObserver<int> obserwator in obserwatorzy)
                obserwator.OnCompleted();
            obserwatorzy.Clear();
        }

        /*
        public void OdłączObserwatora(IObserwator obserwator)
        {
            if (obserwatorzy.Contains(obserwator))
                obserwatorzy.Remove(obserwator);
        }
        */

        public Podmiot()
        {
            wartość = 0;
            //dzielenie = new Dzielenie(3);
            //modulo = new Modulo(3);
        }

        public void ZmieńWartość(int nowaWartość)
        {
            //wartość = nowaWartość;
            //powiadom();

            foreach(IObserver<int> obserwator in obserwatorzy)
            {
                wartość = nowaWartość;
                obserwator.OnNext(wartość);
            }
        }
    }
}